# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Attempt to check each interface in nipype
"""
from __future__ import print_function, unicode_literals
from builtins import object, str, bytes, open

# Stdlib imports
import os
import re
import sys
import warnings
from distutils.version import LooseVersion

from nipype.interfaces.base import BaseInterface

import yapf
if LooseVersion(yapf.__version__) < '0.27':
    raise ImportError("Please upgrade yapf to version 0.27 or newer for stable formatting")

from yapf.yapflib.yapf_api import FormatCode


# Functions and classes
class InterfaceChecker(object):
    """Class for checking all interface specifications
    """

    def __init__(self,
                 package_name,
                 package_skip_patterns=None,
                 module_skip_patterns=None,
                 class_skip_patterns=None):
        r''' Initialize package for parsing

        Parameters
        ----------
        package_name : string
            Name of the top-level package.  *package_name* must be the
            name of an importable package
        package_skip_patterns : None or sequence of {strings, regexps}
            Sequence of strings giving URIs of packages to be excluded
            Operates on the package path, starting at (including) the
            first dot in the package path, after *package_name* - so,
            if *package_name* is ``sphinx``, then ``sphinx.util`` will
            result in ``.util`` being passed for earching by these
            regexps.  If is None, gives default. Default is:
            ['\.tests$']
        module_skip_patterns : None or sequence
            Sequence of strings giving URIs of modules to be excluded
            Operates on the module name including preceding URI path,
            back to the first dot after *package_name*.  For example
            ``sphinx.util.console`` results in the string to search of
            ``.util.console``
            If is None, gives default. Default is:
            ['\.setup$', '\._']
        class_skip_patterns : None or sequence
            Sequence of strings giving classes to be excluded
            Default is: None

        '''
        if package_skip_patterns is None:
            package_skip_patterns = ['\\.tests$']
        if module_skip_patterns is None:
            module_skip_patterns = ['\\.setup$', '\\._']
        if class_skip_patterns:
            self.class_skip_patterns = class_skip_patterns
        else:
            self.class_skip_patterns = []
        self.package_name = package_name
        self.package_skip_patterns = package_skip_patterns
        self.module_skip_patterns = module_skip_patterns

    def get_package_name(self):
        return self._package_name

    def set_package_name(self, package_name):
        """Set package_name"""
        # It's also possible to imagine caching the module parsing here
        self._package_name = package_name
        self.root_module = __import__(package_name)
        self.root_path = self.root_module.__path__[0]

    package_name = property(get_package_name, set_package_name, None,
                            'get/set package_name')

    def _get_object_name(self, line):
        name = line.split()[1].split('(')[0].strip()
        # in case we have classes which are not derived from object
        # ie. old style classes
        return name.rstrip(':')

    def _uri2path(self, uri):
        """Convert uri to absolute filepath

        Parameters
        ----------
        uri : string
            URI of python module to return path for

        Returns
        -------
        path : None or string
            Returns None if there is no valid path for this URI
            Otherwise returns absolute file system path for URI

        """
        if uri == self.package_name:
            return os.path.join(self.root_path, '__init__.py')
        path = uri.replace('.', os.path.sep)
        path = path.replace(self.package_name + os.path.sep, '')
        path = os.path.join(self.root_path, path)
        # XXX maybe check for extensions as well?
        if os.path.exists(path + '.py'):  # file
            path += '.py'
        elif os.path.exists(os.path.join(path, '__init__.py')):
            path = os.path.join(path, '__init__.py')
        else:
            return None
        return path

    def _path2uri(self, dirpath):
        ''' Convert directory path to uri '''
        relpath = dirpath.replace(self.root_path, self.package_name)
        if relpath.startswith(os.path.sep):
            relpath = relpath[1:]
        return relpath.replace(os.path.sep, '.')

    def _parse_module(self, uri):
        ''' Parse module defined in *uri* '''
        filename = self._uri2path(uri)
        if filename is None:
            # nothing that we could handle here.
            return ([], [])
        f = open(filename, 'rt')
        functions, classes = self._parse_lines(f, uri)
        f.close()
        return functions, classes

    def _parse_lines(self, linesource, module):
        ''' Parse lines of text for functions and classes '''
        functions = []
        classes = []
        for line in linesource:
            if line.startswith('def ') and line.count('('):
                # exclude private stuff
                name = self._get_object_name(line)
                if not name.startswith('_'):
                    functions.append(name)
            elif line.startswith('class '):
                # exclude private stuff
                name = self._get_object_name(line)
                if not name.startswith('_') and \
                        self._survives_exclude('.'.join((module, name)),
                                               'class'):
                    classes.append(name)
            else:
                pass
        functions.sort()
        classes.sort()
        return functions, classes

    @classmethod
    def _normalize_repr(cls, value):
        if isinstance(value, list):
            return '[{}]'.format(', '.join(map(cls._normalize_repr, value)))
        if isinstance(value, tuple):
            if len(value) == 1:
                return '({},)'.format(cls._normalize_repr(value[0]))
            return '({})'.format(', '.join(map(cls._normalize_repr, value)))
        if isinstance(value, (str, bytes)):
            value = repr(value)
            if value[0] not in ('"', "'"):
                value = value[1:]
        else:
            value = repr(value)
        return value

    def test_specs(self, uri):
        """Check input and output specs in an uri

        Parameters
        ----------
        uri : string
            python location of module - e.g 'sphinx.builder'

        Returns
        -------
        """
        # get the names of all classes and functions
        _, classes = self._parse_module(uri)
        if not classes:
            # print 'WARNING: Empty -',uri  # dbg
            return None

        # Make a shorter version of the uri that omits the package name for
        # titles
        allowed_keys = [
            'desc', 'genfile', 'xor', 'requires', 'desc', 'nohash', 'argstr',
            'position', 'mandatory', 'copyfile', 'usedefault', 'sep',
            'hash_files', 'deprecated', 'new_name', 'min_ver', 'max_ver',
            'name_source', 'name_template', 'keep_extension', 'units',
            'output_name', 'extensions'
        ]
        in_built = [
            'type', 'copy', 'parent', 'instance_handler', 'comparison_mode',
            'array', 'default', 'editor'
        ]
        bad_specs = []
        for c in classes:
            __import__(uri)
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    classinst = sys.modules[uri].__dict__[c]
            except Exception:
                continue

            if not issubclass(classinst, BaseInterface):
                continue
            testdir = os.path.join(*(uri.split('.')[:-1] + ['tests']))
            if not os.path.exists(testdir):
                os.makedirs(testdir)
            nonautotest = os.path.join(testdir, 'test_%s.py' % c)
            testfile = os.path.join(testdir, 'test_auto_%s.py' % c)
            if os.path.exists(testfile):
                os.unlink(testfile)

            if not os.path.exists(nonautotest):
                cmd = [
                    '# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT',
                    'from __future__ import unicode_literals',
                    'from ..%s import %s' % (uri.split('.')[-1], c), ''
                ]
                cmd.append('\ndef test_%s_inputs():' % c)
                input_fields = ''
                for traitname, trait in sorted(
                        classinst.input_spec().traits(transient=None).items()):
                    input_fields += '%s=dict(' % traitname
                    for key, value in sorted(trait.__dict__.items()):
                        if key in in_built or key == 'desc':
                            continue
                        input_fields += "%s=%s,\n    " % (
                            key, self._normalize_repr(value))
                    input_fields += '),\n    '
                cmd += ['    input_map = dict(%s)' % input_fields]
                cmd += ['    inputs = %s.input_spec()' % c]
                cmd += [
                    """
    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value"""
                ]

                fmt_cmd, _ = FormatCode(
                    '\n'.join(cmd) + '\n\n',
                    style_config={
                        'based_on_style': 'pep8',
                        'column_limit': 79
                    })
                with open(testfile, 'wt') as fp:
                    fp.writelines(fmt_cmd)
            else:
                print('%s has nonautotest' % c)

            for traitname, trait in sorted(
                    classinst.input_spec().traits(transient=None).items()):
                for key in sorted(trait.__dict__):
                    if key in in_built:
                        continue
                    parent_metadata = []
                    if 'parent' in trait.__dict__:
                        parent_metadata = list(
                            getattr(trait, 'parent').__dict__.keys())
                    if key not in allowed_keys + classinst._additional_metadata\
                            + parent_metadata:
                        bad_specs.append([uri, c, 'Inputs', traitname, key])
                    if key == 'mandatory' and trait.mandatory is not None and not trait.mandatory:
                        bad_specs.append(
                            [uri, c, 'Inputs', traitname, 'mandatory=False'])
                    if key == "usedefault" and trait.__dict__[key] == False:
                        bad_specs.append(
                            [uri, c, 'Inputs', traitname, 'usedefault=False'])
                # checking if traits that have default_value different that the trits default one
                # also have `usedefault` specified;
                # excluding TraitCompound
                # excluding Enum: always has default value (the first value)
                # excluding Tuple: takes tuple of inner traits default values as default, but doesn't use it
                # for Range assuming that if default == low, it's likely that usedefault should be False
                # (for Range traits takes low as a default default
                if trait.trait_type.__class__.__name__ not in ['TraitCompound', "Tuple", "Enum"]\
                        and trait.default and "usedefault" not in trait.__dict__\
                        and "requires" not in trait.__dict__\
                        and "xor" not in trait.__dict__:
                    if trait.trait_type.__class__.__name__ is "Range"\
                            and trait.default == trait.trait_type._low:
                        continue
                    bad_specs.append(
                        [uri, c, 'Inputs', traitname, 'default value is set, no value for usedefault'])

            if not classinst.output_spec:
                continue

            if not os.path.exists(nonautotest):
                cmd = ['\ndef test_%s_outputs():' % c]
                input_fields = ''
                for traitname, trait in sorted(classinst.output_spec().traits(
                        transient=None).items()):
                    input_fields += '%s=dict(' % traitname
                    for key, value in sorted(trait.__dict__.items()):
                        if key in in_built or key == 'desc':
                            continue
                        input_fields += "%s=%s,\n    " % (
                            key, self._normalize_repr(value))
                    input_fields += '),\n    '
                cmd += ['    output_map = dict(%s)' % input_fields]
                cmd += ['    outputs = %s.output_spec()' % c]
                cmd += [
                    """
    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value"""
                ]
                fmt_cmd, _ = FormatCode(
                    '\n'.join(cmd) + '\n',
                    style_config={
                        'based_on_style': 'pep8',
                        'column_limit': 79
                    })
                with open(testfile, 'at') as fp:
                    fp.writelines(fmt_cmd)

            for traitname, trait in sorted(
                    classinst.output_spec().traits(transient=None).items()):
                for key in sorted(trait.__dict__):
                    if key in in_built:
                        continue
                    parent_metadata = []
                    if 'parent' in trait.__dict__:
                        parent_metadata = list(
                            getattr(trait, 'parent').__dict__.keys())
                    if key not in allowed_keys + classinst._additional_metadata\
                            + parent_metadata:
                        bad_specs.append([uri, c, 'Outputs', traitname, key])
        return bad_specs

    def _survives_exclude(self, matchstr, match_type):
        ''' Returns True if *matchstr* does not match patterns

        ``self.package_name`` removed from front of string if present

        Examples
        --------
        >>> dw = ApiDocWriter('sphinx')
        >>> dw._survives_exclude('sphinx.okpkg', 'package')
        True
        >>> dw.package_skip_patterns.append('^\\.badpkg$')
        >>> dw._survives_exclude('sphinx.badpkg', 'package')
        False
        >>> dw._survives_exclude('sphinx.badpkg', 'module')
        True
        >>> dw._survives_exclude('sphinx.badmod', 'module')
        True
        >>> dw.module_skip_patterns.append('^\\.badmod$')
        >>> dw._survives_exclude('sphinx.badmod', 'module')
        False
        '''
        if match_type == 'module':
            patterns = self.module_skip_patterns
        elif match_type == 'package':
            patterns = self.package_skip_patterns
        elif match_type == 'class':
            patterns = self.class_skip_patterns
        else:
            raise ValueError('Cannot interpret match type "%s"' % match_type)
        # Match to URI without package name
        L = len(self.package_name)
        if matchstr[:L] == self.package_name:
            matchstr = matchstr[L:]
        for pat in patterns:
            try:
                pat.search
            except AttributeError:
                pat = re.compile(pat)
            if pat.search(matchstr):
                return False
        return True

    def discover_modules(self):
        ''' Return module sequence discovered from ``self.package_name``


        Parameters
        ----------
        None

        Returns
        -------
        mods : sequence
            Sequence of module names within ``self.package_name``

        Examples
        --------
        '''
        modules = [self.package_name]
        # raw directory parsing
        for dirpath, dirnames, filenames in os.walk(self.root_path):
            # Check directory names for packages
            root_uri = self._path2uri(os.path.join(self.root_path, dirpath))
            for dirname in dirnames[:]:  # copy list - we modify inplace
                package_uri = '.'.join((root_uri, dirname))
                if (self._uri2path(package_uri)
                        and self._survives_exclude(package_uri, 'package')):
                    modules.append(package_uri)
                else:
                    dirnames.remove(dirname)
            # Check filenames for modules
            for filename in filenames:
                module_name = filename[:-3]
                module_uri = '.'.join((root_uri, module_name))
                if (self._uri2path(module_uri)
                        and self._survives_exclude(module_uri, 'module')):
                    modules.append(module_uri)
        return sorted(modules)

    def check_modules(self):
        # write the list
        modules = self.discover_modules()
        checked_modules = []
        for m in modules:
            bad_specs = self.test_specs(m)
            if bad_specs:
                checked_modules.extend(bad_specs)
        for bad_spec in checked_modules:
            print(':'.join(bad_spec))


if __name__ == "__main__":
    package = 'nipype'
    ic = InterfaceChecker(package)
    # Packages that should not be included in generated API docs.
    ic.package_skip_patterns += [
        '\.external$',
        '\.fixes$',
        '\.utils$',
        '\.pipeline',
        '\.testing',
        '\.caching',
        '\.workflows',
    ]
    """
    # Modules that should not be included in generated API docs.
    ic.module_skip_patterns += ['\.version$',
                                '\.interfaces\.base$',
                                '\.interfaces\.matlab$',
                                '\.interfaces\.rest$',
                                '\.interfaces\.pymvpa$',
                                '\.interfaces\.slicer\.generate_classes$',
                                '\.interfaces\.spm\.base$',
                                '\.interfaces\.traits',
                                '\.pipeline\.alloy$',
                                '\.pipeline\.s3_node_wrapper$',
                                '.\testing',
                                       ]
    ic.class_skip_patterns += ['AFNI',
                               'ANTS',
                               'FSL',
                               'FS',
                               'Info',
                               '^SPM',
                               'Tester',
                               'Spec$',
                               'Numpy',
                               'NipypeTester',
                                      ]
    """
    ic.check_modules()



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..segmentation import JointFusion


def test_JointFusion_inputs():
    input_map = dict(
        alpha=dict(
            requires=['method'],
            usedefault=True,
        ),
        args=dict(argstr='%s', ),
        atlas_group_id=dict(argstr='-gp %d...', ),
        atlas_group_weights=dict(argstr='-gpw %d...', ),
        beta=dict(
            requires=['method'],
            usedefault=True,
        ),
        dimension=dict(
            argstr='%d',
            mandatory=True,
            position=0,
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        exclusion_region=dict(
            argstr='-x %s',
            extensions=None,
        ),
        method=dict(
            argstr='-m %s',
            usedefault=True,
        ),
        modalities=dict(
            argstr='%d',
            mandatory=True,
            position=1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        output_label_image=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            name_template='%s',
            output_name='output_label_image',
            position=-1,
        ),
        patch_radius=dict(
            argstr='-rp %s',
            maxlen=3,
            minlen=3,
        ),
        search_radius=dict(
            argstr='-rs %s',
            maxlen=3,
            minlen=3,
        ),
        target_image=dict(
            argstr='-tg %s...',
            mandatory=True,
        ),
        warped_intensity_images=dict(
            argstr='-g %s...',
            mandatory=True,
        ),
        warped_label_images=dict(
            argstr='-l %s...',
            mandatory=True,
        ),
    )
    inputs = JointFusion.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_JointFusion_outputs():
    output_map = dict(output_label_image=dict(extensions=None, ), )
    outputs = JointFusion.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import FitMSParams


def test_FitMSParams_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flip_list=dict(),
        in_files=dict(
            argstr='%s',
            mandatory=True,
            position=-2,
        ),
        out_dir=dict(
            argstr='%s',
            genfile=True,
            position=-1,
        ),
        subjects_dir=dict(),
        te_list=dict(),
        tr_list=dict(),
        xfm_list=dict(),
    )
    inputs = FitMSParams.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_FitMSParams_outputs():
    output_map = dict(
        pd_image=dict(extensions=None, ),
        t1_image=dict(extensions=None, ),
        t2star_image=dict(extensions=None, ),
    )
    outputs = FitMSParams.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..specialized import EMSegmentCommandLine


def test_EMSegmentCommandLine_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        atlasVolumeFileNames=dict(argstr='--atlasVolumeFileNames %s...', ),
        disableCompression=dict(argstr='--disableCompression ', ),
        disableMultithreading=dict(argstr='--disableMultithreading %d', ),
        dontUpdateIntermediateData=dict(
            argstr='--dontUpdateIntermediateData %d', ),
        dontWriteResults=dict(argstr='--dontWriteResults ', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        generateEmptyMRMLSceneAndQuit=dict(
            argstr='--generateEmptyMRMLSceneAndQuit %s',
            hash_files=False,
        ),
        intermediateResultsDirectory=dict(
            argstr='--intermediateResultsDirectory %s', ),
        keepTempFiles=dict(argstr='--keepTempFiles ', ),
        loadAtlasNonCentered=dict(argstr='--loadAtlasNonCentered ', ),
        loadTargetCentered=dict(argstr='--loadTargetCentered ', ),
        mrmlSceneFileName=dict(
            argstr='--mrmlSceneFileName %s',
            extensions=None,
        ),
        parametersMRMLNodeName=dict(argstr='--parametersMRMLNodeName %s', ),
        registrationAffineType=dict(argstr='--registrationAffineType %d', ),
        registrationDeformableType=dict(
            argstr='--registrationDeformableType %d', ),
        registrationPackage=dict(argstr='--registrationPackage %s', ),
        resultMRMLSceneFileName=dict(
            argstr='--resultMRMLSceneFileName %s',
            hash_files=False,
        ),
        resultStandardVolumeFileName=dict(
            argstr='--resultStandardVolumeFileName %s',
            extensions=None,
        ),
        resultVolumeFileName=dict(
            argstr='--resultVolumeFileName %s',
            hash_files=False,
        ),
        targetVolumeFileNames=dict(argstr='--targetVolumeFileNames %s...', ),
        taskPreProcessingSetting=dict(
            argstr='--taskPreProcessingSetting %s', ),
        verbose=dict(argstr='--verbose ', ),
    )
    inputs = EMSegmentCommandLine.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_EMSegmentCommandLine_outputs():
    output_map = dict(
        generateEmptyMRMLSceneAndQuit=dict(extensions=None, ),
        resultMRMLSceneFileName=dict(extensions=None, ),
        resultVolumeFileName=dict(extensions=None, ),
    )
    outputs = EMSegmentCommandLine.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..regutils import RegResample


def test_RegResample_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flo_file=dict(
            argstr='-flo %s',
            extensions=None,
            mandatory=True,
        ),
        inter_val=dict(argstr='-inter %d', ),
        omp_core_val=dict(
            argstr='-omp %i',
            usedefault=True,
        ),
        out_file=dict(
            argstr='%s',
            extensions=None,
            name_source=['flo_file'],
            name_template='%s',
            position=-1,
        ),
        pad_val=dict(argstr='-pad %f', ),
        psf_alg=dict(argstr='-psf_alg %d', ),
        psf_flag=dict(argstr='-psf', ),
        ref_file=dict(
            argstr='-ref %s',
            extensions=None,
            mandatory=True,
        ),
        tensor_flag=dict(argstr='-tensor ', ),
        trans_file=dict(
            argstr='-trans %s',
            extensions=None,
        ),
        type=dict(
            argstr='-%s',
            position=-2,
            usedefault=True,
        ),
        verbosity_off_flag=dict(argstr='-voff', ),
    )
    inputs = RegResample.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_RegResample_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = RegResample.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os
import warnings
import xml.dom.minidom

from .base import (CommandLine, CommandLineInputSpec, DynamicTraitedSpec,
                   traits, Undefined, File, isdefined)


class SlicerCommandLineInputSpec(DynamicTraitedSpec, CommandLineInputSpec):
    module = traits.Str(
        desc="name of the Slicer command line module you want to use")


class SlicerCommandLine(CommandLine):
    """Experimental Slicer wrapper. Work in progress.

    """
    _cmd = "Slicer3"
    input_spec = SlicerCommandLineInputSpec
    output_spec = DynamicTraitedSpec

    def _grab_xml(self, module):
        cmd = CommandLine(
            command="Slicer3",
            resource_monitor=False,
            args="--launch %s --xml" % module)
        ret = cmd.run()
        if ret.runtime.returncode == 0:
            return xml.dom.minidom.parseString(ret.runtime.stdout)
        else:
            raise Exception(cmd.cmdline + " failed:\n%s" % ret.runtime.stderr)

    def _outputs(self):
        base = super(SlicerCommandLine, self)._outputs()
        undefined_output_traits = {}
        for key in [
                node.getElementsByTagName('name')[0].firstChild.nodeValue
                for node in self._outputs_nodes
        ]:
            base.add_trait(key, File(exists=True))
            undefined_output_traits[key] = Undefined

        base.trait_set(trait_change_notify=False, **undefined_output_traits)
        return base

    def __init__(self, module, **inputs):
        warnings.warn('slicer is Not fully implemented', RuntimeWarning)
        super(SlicerCommandLine, self).__init__(
            command="Slicer3 --launch %s " % module, name=module, **inputs)
        dom = self._grab_xml(module)
        self._outputs_filenames = {}

        self._outputs_nodes = []

        undefined_traits = {}

        for paramGroup in dom.getElementsByTagName("parameters"):
            for param in paramGroup.childNodes:
                if param.nodeName in [
                        'label', 'description', '#text', '#comment'
                ]:
                    continue
                traitsParams = {}

                name = param.getElementsByTagName('name')[
                    0].firstChild.nodeValue

                longFlagNode = param.getElementsByTagName('longflag')
                if longFlagNode:
                    traitsParams[
                        "argstr"] = "--" + longFlagNode[0].firstChild.nodeValue + " "
                else:
                    traitsParams["argstr"] = "--" + name + " "

                argsDict = {
                    'file': '%s',
                    'integer': "%d",
                    'double': "%f",
                    'float': "%f",
                    'image': "%s",
                    'transform': "%s",
                    'boolean': '',
                    'string-enumeration': '%s',
                    'string': "%s"
                }

                if param.nodeName.endswith('-vector'):
                    traitsParams["argstr"] += argsDict[param.nodeName[:-7]]
                else:
                    traitsParams["argstr"] += argsDict[param.nodeName]

                index = param.getElementsByTagName('index')
                if index:
                    traitsParams["position"] = index[0].firstChild.nodeValue

                desc = param.getElementsByTagName('description')
                if index:
                    traitsParams["desc"] = desc[0].firstChild.nodeValue

                name = param.getElementsByTagName('name')[
                    0].firstChild.nodeValue

                typesDict = {
                    'integer': traits.Int,
                    'double': traits.Float,
                    'float': traits.Float,
                    'image': File,
                    'transform': File,
                    'boolean': traits.Bool,
                    'string': traits.Str,
                    'file': File
                }

                if param.nodeName == 'string-enumeration':
                    type = traits.Enum
                    values = [
                        el.firstChild.nodeValue
                        for el in param.getElementsByTagName('element')
                    ]
                elif param.nodeName.endswith('-vector'):
                    type = traits.List
                    values = [typesDict[param.nodeName[:-7]]]
                    traitsParams["sep"] = ','
                else:
                    values = []
                    type = typesDict[param.nodeName]

                if param.nodeName in [
                        'file', 'directory', 'image', 'transform'
                ] and param.getElementsByTagName(
                        'channel')[0].firstChild.nodeValue == 'output':
                    self.inputs.add_trait(name,
                                          traits.Either(
                                              traits.Bool, File,
                                              **traitsParams))
                    undefined_traits[name] = Undefined

                    # traitsParams["exists"] = True
                    self._outputs_filenames[
                        name] = self._gen_filename_from_param(param)
                    # undefined_output_traits[name] = Undefined
                    # self._outputs().add_trait(name, File(*values, **traitsParams))
                    self._outputs_nodes.append(param)
                else:
                    if param.nodeName in [
                            'file', 'directory', 'image', 'transform'
                    ]:
                        traitsParams["exists"] = True
                    self.inputs.add_trait(name, type(*values, **traitsParams))
                    undefined_traits[name] = Undefined

        self.inputs.trait_set(trait_change_notify=False, **undefined_traits)
        for name in list(undefined_traits.keys()):
            _ = getattr(self.inputs, name)
        # self._outputs().trait_set(trait_change_notify=False, **undefined_output_traits)

    def _gen_filename(self, name):
        if name in self._outputs_filenames:
            return os.path.join(os.getcwd(), self._outputs_filenames[name])
        return None

    def _gen_filename_from_param(self, param):
        base = param.getElementsByTagName('name')[0].firstChild.nodeValue
        fileExtensions = param.getAttribute("fileExtensions")
        if fileExtensions:
            ext = fileExtensions
        else:
            ext = {
                'image': '.nii',
                'transform': '.txt',
                'file': ''
            }[param.nodeName]
        return base + ext

    def _list_outputs(self):
        outputs = self.output_spec().get()
        for output_node in self._outputs_nodes:
            name = output_node.getElementsByTagName('name')[
                0].firstChild.nodeValue
            outputs[name] = getattr(self.inputs, name)
            if isdefined(outputs[name]) and isinstance(outputs[name], bool):
                if outputs[name]:
                    outputs[name] = self._gen_filename(name)
                else:
                    outputs[name] = Undefined
        return outputs

    def _format_arg(self, name, spec, value):
        if name in [
                output_node.getElementsByTagName('name')[0]
                .firstChild.nodeValue for output_node in self._outputs_nodes
        ]:
            if isinstance(value, bool):
                fname = self._gen_filename(name)
            else:
                fname = value
            return spec.argstr % fname
        return super(SlicerCommandLine, self)._format_arg(name, spec, value)


#    test = SlicerCommandLine(module="BRAINSFit")
#    test.inputs.fixedVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/10_co_COR_3D_IR_PREP.nii"
#    test.inputs.movingVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/2_line_bisection.nii"
#    test.inputs.outputTransform = True
#    test.inputs.transformType = ["Affine"]
#    print test.cmdline
#    print test.inputs
#    print test._outputs()
#    ret = test.run()

#    test = SlicerCommandLine(name="BRAINSResample")
#    test.inputs.referenceVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/10_co_COR_3D_IR_PREP.nii"
#    test.inputs.inputVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/2_line_bisection.nii"
#    test.inputs.outputVolume = True
#    test.inputs.warpTransform = "/home/filo/workspace/nipype/nipype/interfaces/outputTransform.mat"
#    print test.cmdline
#    ret = test.run()
#    print ret.runtime.stderr
#    print ret.runtime.returncode



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..registration import DiffeoSymTensor3DVol


def test_DiffeoSymTensor3DVol_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        df=dict(
            argstr='-df %s',
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flip=dict(argstr='-flip %d %d %d', ),
        in_file=dict(
            argstr='-in %s',
            extensions=None,
            mandatory=True,
        ),
        interpolation=dict(
            argstr='-interp %s',
            usedefault=True,
        ),
        out_file=dict(
            argstr='-out %s',
            extensions=None,
            keep_extension=True,
            name_source='in_file',
            name_template='%s_diffeoxfmd',
        ),
        reorient=dict(
            argstr='-reorient %s',
            usedefault=True,
        ),
        resampling_type=dict(argstr='-type %s', ),
        target=dict(
            argstr='-target %s',
            extensions=None,
            xor=['voxel_size'],
        ),
        transform=dict(
            argstr='-trans %s',
            extensions=None,
            mandatory=True,
        ),
        voxel_size=dict(
            argstr='-vsize %g %g %g',
            xor=['target'],
        ),
    )
    inputs = DiffeoSymTensor3DVol.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_DiffeoSymTensor3DVol_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = DiffeoSymTensor3DVol.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
""" Base interfaces for dipy """
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os.path as op
import inspect
import numpy as np
from ... import logging
from ..base import (traits, File, isdefined, LibraryBaseInterface,
                    BaseInterfaceInputSpec, TraitedSpec)

# List of workflows to ignore
SKIP_WORKFLOWS_LIST = ['Workflow', 'CombinedWorkflow']

HAVE_DIPY = True

try:
    import dipy
    from dipy.workflows.base import IntrospectiveArgumentParser
except ImportError:
    HAVE_DIPY = False


def no_dipy():
    """ Check if dipy is available """
    global HAVE_DIPY
    return not HAVE_DIPY


def dipy_version():
    """ Check dipy version """
    if no_dipy():
        return None

    return dipy.__version__


class DipyBaseInterface(LibraryBaseInterface):
    """
    A base interface for py:mod:`dipy` computations
    """
    _pkg = 'dipy'


class DipyBaseInterfaceInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc=('input diffusion data'))
    in_bval = File(exists=True, mandatory=True, desc=('input b-values table'))
    in_bvec = File(exists=True, mandatory=True, desc=('input b-vectors table'))
    b0_thres = traits.Int(700, usedefault=True, desc=('b0 threshold'))
    out_prefix = traits.Str(desc=('output prefix for file names'))


class DipyDiffusionInterface(DipyBaseInterface):
    """
    A base interface for py:mod:`dipy` computations
    """
    input_spec = DipyBaseInterfaceInputSpec

    def _get_gradient_table(self):
        bval = np.loadtxt(self.inputs.in_bval)
        bvec = np.loadtxt(self.inputs.in_bvec).T
        from dipy.core.gradients import gradient_table
        gtab = gradient_table(bval, bvec)

        gtab.b0_threshold = self.inputs.b0_thres
        return gtab

    def _gen_filename(self, name, ext=None):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == '.gz':
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext

        if not isdefined(self.inputs.out_prefix):
            out_prefix = op.abspath(fname)
        else:
            out_prefix = self.inputs.out_prefix

        if ext is None:
            ext = fext

        return out_prefix + '_' + name + ext


def convert_to_traits_type(dipy_type, is_file=False):
    """Convert DIPY type to Traits type."""
    dipy_type = dipy_type.lower()
    is_mandatory = bool("optional" not in dipy_type)
    if "variable" in dipy_type and "string" in dipy_type:
        return traits.ListStr, is_mandatory
    elif "variable" in dipy_type and "int" in dipy_type:
        return traits.ListInt, is_mandatory
    elif "variable" in dipy_type and "float" in dipy_type:
        return traits.ListFloat, is_mandatory
    elif "variable" in dipy_type and "bool" in dipy_type:
        return traits.ListBool, is_mandatory
    elif "variable" in dipy_type and "complex" in dipy_type:
        return traits.ListComplex, is_mandatory
    elif "string" in dipy_type and not is_file:
        return traits.Str, is_mandatory
    elif "string" in dipy_type and is_file:
        return File, is_mandatory
    elif "int" in dipy_type:
        return traits.Int, is_mandatory
    elif "float" in dipy_type:
        return traits.Float, is_mandatory
    elif "bool" in dipy_type:
        return traits.Bool, is_mandatory
    elif "complex" in dipy_type:
        return traits.Complex, is_mandatory
    else:
        msg = "Error during convert_to_traits_type({0}).".format(dipy_type) + \
              "Unknown DIPY type."
        raise IOError(msg)


def create_interface_specs(class_name, params=None, BaseClass=TraitedSpec):
    """Create IN/Out interface specifications dynamically.

    Parameters
    ----------
    class_name: str
        The future class name(e.g, (MyClassInSpec))
    params: list of tuple
        dipy argument list
    BaseClass: TraitedSpec object
        parent class

    Returns
    -------
    newclass: object
        new nipype interface specification class

    """
    attr = {}
    if params is not None:
        for p in params:
            name, dipy_type, desc = p[0], p[1], p[2]
            is_file = bool("files" in name or "out_" in name)
            traits_type, is_mandatory = convert_to_traits_type(dipy_type,
                                                               is_file)
            # print(name, dipy_type, desc, is_file, traits_type, is_mandatory)
            if BaseClass.__name__ == BaseInterfaceInputSpec.__name__:
                if len(p) > 3:
                    attr[name] = traits_type(p[3], desc=desc[-1],
                                             usedefault=True,
                                             mandatory=is_mandatory)
                else:
                    attr[name] = traits_type(desc=desc[-1],
                                             mandatory=is_mandatory)
            else:
                attr[name] = traits_type(p[3], desc=desc[-1], exists=True,
                                         usedefault=True,)

    newclass = type(str(class_name), (BaseClass, ), attr)
    return newclass


def dipy_to_nipype_interface(cls_name, dipy_flow, BaseClass=DipyBaseInterface):
    """Construct a class in order to respect nipype interface specifications.

    This convenient class factory convert a DIPY Workflow to a nipype
    interface.

    Parameters
    ----------
    cls_name: string
        new class name
    dipy_flow: Workflow class type.
        It should be any children class of `dipy.workflows.workflow.Worflow`
    BaseClass: object
        nipype instance object

    Returns
    -------
    newclass: object
        new nipype interface specification class

    """
    parser = IntrospectiveArgumentParser()
    flow = dipy_flow()
    parser.add_workflow(flow)
    default_values = inspect.getargspec(flow.run).defaults
    optional_params = [args + (val,) for args, val in zip(parser.optional_parameters, default_values)]
    start = len(parser.optional_parameters) - len(parser.output_parameters)

    output_parameters = [args + (val,) for args, val in zip(parser.output_parameters, default_values[start:])]
    input_parameters = parser.positional_parameters + optional_params

    input_spec = create_interface_specs("{}InputSpec".format(cls_name),
                                        input_parameters,
                                        BaseClass=BaseInterfaceInputSpec)

    output_spec = create_interface_specs("{}OutputSpec".format(cls_name),
                                         output_parameters,
                                         BaseClass=TraitedSpec)

    def _run_interface(self, runtime):
        flow = dipy_flow()
        args = self.inputs.get()
        flow.run(**args)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_dir = outputs.get("out_dir", ".")
        for key, values in outputs.items():
            outputs[key] = op.join(out_dir, values)

        return outputs

    newclass = type(str(cls_name), (BaseClass, ),
                    {"input_spec": input_spec,
                     "output_spec": output_spec,
                     "_run_interface": _run_interface,
                     "_list_outputs:": _list_outputs})
    return newclass


def get_dipy_workflows(module):
    """Search for DIPY workflow class.

    Parameters
    ----------
    module : object
        module object

    Returns
    -------
    l_wkflw : list of tuple
        This a list of tuple containing 2 elements:
        Worflow name, Workflow class obj

    Examples
    --------
    >>> from dipy.workflows import align  # doctest: +SKIP
    >>> get_dipy_workflows(align)  # doctest: +SKIP

    """
    return [(m, obj) for m, obj in inspect.getmembers(module)
            if inspect.isclass(obj) and
            issubclass(obj, module.Workflow) and
            m not in SKIP_WORKFLOWS_LIST]



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.

Top-level namespace for fsl.
"""

from .base import (FSLCommand, Info, check_fsl, no_fsl, no_fsl_course_data)
from .preprocess import (FAST, FLIRT, ApplyXFM, BET, MCFLIRT, FNIRT, ApplyWarp,
                         SliceTimer, SUSAN, PRELUDE, FUGUE, FIRST)
from .model import (Level1Design, FEAT, FEATModel, FILMGLS, FEATRegister,
                    FLAMEO, ContrastMgr, MultipleRegressDesign, L2Model, SMM,
                    MELODIC, SmoothEstimate, Cluster, Randomise, GLM)
from .utils import (
    AvScale, Smooth, Slice, Merge, ExtractROI, Split, ImageMaths, ImageMeants,
    ImageStats, FilterRegressor, Overlay, Slicer, PlotTimeSeries,
    PlotMotionParams, ConvertXFM, SwapDimensions, PowerSpectrum, Reorient2Std,
    Complex, InvWarp, WarpUtils, ConvertWarp, WarpPoints, WarpPointsToStd,
    WarpPointsFromStd, RobustFOV, CopyGeom, MotionOutliers)

from .epi import (PrepareFieldmap, TOPUP, ApplyTOPUP, Eddy, EPIDeWarp, SigLoss,
                  EddyCorrect, EpiReg, EddyQuad)
from .dti import (BEDPOSTX, XFibres, DTIFit, ProbTrackX, ProbTrackX2, VecReg,
                  ProjThresh, FindTheBiggest, DistanceMap, TractSkeleton,
                  MakeDyadicVectors, BEDPOSTX5, XFibres5)
from .maths import (ChangeDataType, Threshold, MeanImage, ApplyMask,
                    IsotropicSmooth, TemporalFilter, DilateImage, ErodeImage,
                    SpatialFilter, UnaryMaths, BinaryMaths, MultiImageMaths,
                    MaxnImage, MinImage, MedianImage, PercentileImage,
                    AR1Image)
from .possum import B0Calc
from .fix import (AccuracyTester, Classifier, Cleaner, FeatureExtractor,
                  Training, TrainingSetCreator)
from .aroma import ICA_AROMA



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function, unicode_literals
import os
import pytest

from nipype.interfaces import utility
from nipype.interfaces.base import isdefined
import nipype.pipeline.engine as pe


def test_rename(tmpdir):
    tmpdir.chdir()

    # Test very simple rename
    _ = open("file.txt", "w").close()
    rn = utility.Rename(in_file="file.txt", format_string="test_file1.txt")
    res = rn.run()
    outfile = tmpdir.join("test_file1.txt").strpath
    assert res.outputs.out_file == outfile
    assert os.path.exists(outfile)

    # Now a string-formatting version
    rn = utility.Rename(
        in_file="file.txt",
        format_string="%(field1)s_file%(field2)d",
        keep_ext=True)
    # Test .input field creation
    assert hasattr(rn.inputs, "field1")
    assert hasattr(rn.inputs, "field2")

    # Set the inputs
    rn.inputs.field1 = "test"
    rn.inputs.field2 = 2
    res = rn.run()
    outfile = tmpdir.join("test_file2.txt").strpath
    assert res.outputs.out_file == outfile
    assert os.path.exists(outfile)


@pytest.mark.parametrize("args, expected", [({}, ([0], [1, 2, 3])),
                                            ({
                                                "squeeze": True
                                            }, (0, [1, 2, 3]))])
def test_split(tmpdir, args, expected):
    tmpdir.chdir()

    node = pe.Node(
        utility.Split(inlist=list(range(4)), splits=[1, 3], **args),
        name='split_squeeze')
    res = node.run()
    assert res.outputs.out1 == expected[0]
    assert res.outputs.out2 == expected[1]


@pytest.mark.parametrize("args, kwargs, in_lists, expected", [
    ([3], {}, [0, [1, 2], [3, 4, 5]], [0, 1, 2, 3, 4, 5]),
    ([0], {}, None, None),
    ([], {}, [], []),
    ([], {}, [0, [1, 2], [3, 4, 5]], [0, [1, 2], [3, 4, 5]]),
    ([3], {
        'axis': 'hstack'
    }, [[0], [1, 2], [3, 4, 5]], [[0, 1, 3]]),
    ([3], {
        'axis': 'hstack'
    }, [[0, 1], [2, 3], [4, 5]], [[0, 2, 4], [1, 3, 5]]),
    ([3], {
        'axis': 'hstack'
    }, [[0, 1], [2, 3], [4, 5]], [[0, 2, 4], [1, 3, 5]]),
])
def test_merge(tmpdir, args, kwargs, in_lists, expected):
    tmpdir.chdir()

    node = pe.Node(utility.Merge(*args, **kwargs), name='merge')

    numinputs = args[0] if args else 0
    if numinputs >= 1:
        for i in range(1, numinputs + 1):
            setattr(node.inputs, 'in{:d}'.format(i), in_lists[i - 1])

    res = node.run()
    if numinputs < 1:
        assert not isdefined(res.outputs.out)
    else:
        assert res.outputs.out == expected



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import pytest

from ....utils.filemanip import which
from ....testing import example_data
from .. import (get_custom_path, RegAverage, RegResample, RegJacobian,
                RegTools, RegMeasure, RegTransform)


def no_nifty_tool(cmd=None):
    return which(cmd) is None


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_resample'),
    reason="niftyreg is not installed. reg_resample not found.")
def test_reg_resample_res():
    """ tests for reg_resample interface """
    # Create a reg_resample object
    nr_resample = RegResample()

    # Check if the command is properly defined
    assert nr_resample.cmd == get_custom_path('reg_resample')

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_resample.run()

    # Resample res
    ref_file = example_data('im1.nii')
    flo_file = example_data('im2.nii')
    trans_file = example_data('warpfield.nii')
    nr_resample.inputs.ref_file = ref_file
    nr_resample.inputs.flo_file = flo_file
    nr_resample.inputs.trans_file = trans_file
    nr_resample.inputs.inter_val = 'LIN'
    nr_resample.inputs.omp_core_val = 4

    cmd_tmp = '{cmd} -flo {flo} -inter 1 -omp 4 -ref {ref} -trans {trans} \
-res {res}'

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_resample'),
        flo=flo_file,
        ref=ref_file,
        trans=trans_file,
        res='im2_res.nii.gz')

    assert nr_resample.cmdline == expected_cmd

    # test_reg_resample_blank()
    nr_resample_2 = RegResample(type='blank', inter_val='LIN', omp_core_val=4)
    ref_file = example_data('im1.nii')
    flo_file = example_data('im2.nii')
    trans_file = example_data('warpfield.nii')
    nr_resample_2.inputs.ref_file = ref_file
    nr_resample_2.inputs.flo_file = flo_file
    nr_resample_2.inputs.trans_file = trans_file

    cmd_tmp = '{cmd} -flo {flo} -inter 1 -omp 4 -ref {ref} -trans {trans} \
-blank {blank}'

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_resample'),
        flo=flo_file,
        ref=ref_file,
        trans=trans_file,
        blank='im2_blank.nii.gz')

    assert nr_resample_2.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_jacobian'),
    reason="niftyreg is not installed. reg_jacobian not found.")
def test_reg_jacobian_jac():
    """ Test interface for RegJacobian """
    # Create a reg_jacobian object
    nr_jacobian = RegJacobian()

    # Check if the command is properly defined
    assert nr_jacobian.cmd == get_custom_path('reg_jacobian')

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_jacobian.run()

    # Test Reg Jacobian: jac
    ref_file = example_data('im1.nii')
    trans_file = example_data('warpfield.nii')
    nr_jacobian.inputs.ref_file = ref_file
    nr_jacobian.inputs.trans_file = trans_file
    nr_jacobian.inputs.omp_core_val = 4

    cmd_tmp = '{cmd} -omp 4 -ref {ref} -trans {trans} -jac {jac}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_jacobian'),
        ref=ref_file,
        trans=trans_file,
        jac='warpfield_jac.nii.gz')

    assert nr_jacobian.cmdline == expected_cmd

    # Test Reg Jacobian: jac m
    nr_jacobian_2 = RegJacobian(type='jacM', omp_core_val=4)
    ref_file = example_data('im1.nii')
    trans_file = example_data('warpfield.nii')
    nr_jacobian_2.inputs.ref_file = ref_file
    nr_jacobian_2.inputs.trans_file = trans_file

    cmd_tmp = '{cmd} -omp 4 -ref {ref} -trans {trans} -jacM {jac}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_jacobian'),
        ref=ref_file,
        trans=trans_file,
        jac='warpfield_jacM.nii.gz')

    assert nr_jacobian_2.cmdline == expected_cmd

    # Test Reg Jacobian: jac l
    nr_jacobian_3 = RegJacobian(type='jacL', omp_core_val=4)
    ref_file = example_data('im1.nii')
    trans_file = example_data('warpfield.nii')
    nr_jacobian_3.inputs.ref_file = ref_file
    nr_jacobian_3.inputs.trans_file = trans_file

    cmd_tmp = '{cmd} -omp 4 -ref {ref} -trans {trans} -jacL {jac}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_jacobian'),
        ref=ref_file,
        trans=trans_file,
        jac='warpfield_jacL.nii.gz')

    assert nr_jacobian_3.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_tools'),
    reason="niftyreg is not installed. reg_tools not found.")
def test_reg_tools_mul():
    """ tests for reg_tools interface """
    # Create a reg_tools object
    nr_tools = RegTools()

    # Check if the command is properly defined
    assert nr_tools.cmd == get_custom_path('reg_tools')

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_tools.run()

    # Test reg_tools: mul
    in_file = example_data('im1.nii')
    nr_tools.inputs.in_file = in_file
    nr_tools.inputs.mul_val = 4
    nr_tools.inputs.omp_core_val = 4

    cmd_tmp = '{cmd} -in {in_file} -mul 4.0 -omp 4 -out {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_tools'),
        in_file=in_file,
        out_file='im1_tools.nii.gz')

    assert nr_tools.cmdline == expected_cmd

    # Test reg_tools: iso
    nr_tools_2 = RegTools(iso_flag=True, omp_core_val=4)
    in_file = example_data('im1.nii')
    nr_tools_2.inputs.in_file = in_file

    cmd_tmp = '{cmd} -in {in_file} -iso -omp 4 -out {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_tools'),
        in_file=in_file,
        out_file='im1_tools.nii.gz')

    assert nr_tools_2.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_average'),
    reason="niftyreg is not installed. reg_average not found.")
def test_reg_average():
    """ tests for reg_average interface """
    # Create a reg_average object
    nr_average = RegAverage()

    # Check if the command is properly defined
    assert nr_average.cmd == get_custom_path('reg_average')

    # Average niis
    one_file = example_data('im1.nii')
    two_file = example_data('im2.nii')
    three_file = example_data('im3.nii')
    nr_average.inputs.avg_files = [one_file, two_file, three_file]
    nr_average.inputs.omp_core_val = 1
    generated_cmd = nr_average.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), 'reg_average_cmd')
    with open(reg_average_cmd, 'rb') as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = '%s %s -avg %s %s %s -omp 1' % (
        get_custom_path('reg_average'),
        os.path.join(os.getcwd(), 'avg_out.nii.gz'), one_file, two_file,
        three_file)

    assert argv.decode('utf-8') == expected_argv

    # Test command line with text file
    expected_cmd = ('%s --cmd_file %s' % (get_custom_path('reg_average'),
                                          reg_average_cmd))

    assert generated_cmd == expected_cmd

    # Test Reg Average: average txt
    nr_average_2 = RegAverage()
    one_file = example_data('TransformParameters.0.txt')
    two_file = example_data('ants_Affine.txt')
    three_file = example_data('elastix.txt')
    nr_average_2.inputs.avg_files = [one_file, two_file, three_file]
    nr_average_2.inputs.omp_core_val = 1
    generated_cmd = nr_average_2.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), 'reg_average_cmd')
    with open(reg_average_cmd, 'rb') as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = '%s %s -avg %s %s %s -omp 1' % (
        get_custom_path('reg_average'),
        os.path.join(os.getcwd(), 'avg_out.txt'), one_file, two_file,
        three_file)

    assert argv.decode('utf-8') == expected_argv

    # Test Reg Average: average list
    nr_average_3 = RegAverage()
    one_file = example_data('TransformParameters.0.txt')
    two_file = example_data('ants_Affine.txt')
    three_file = example_data('elastix.txt')
    nr_average_3.inputs.avg_lts_files = [one_file, two_file, three_file]
    nr_average_3.inputs.omp_core_val = 1
    generated_cmd = nr_average_3.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), 'reg_average_cmd')
    with open(reg_average_cmd, 'rb') as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = ('%s %s -avg_lts %s %s %s -omp 1' %
                     (get_custom_path('reg_average'),
                      os.path.join(os.getcwd(), 'avg_out.txt'), one_file,
                      two_file, three_file))

    assert argv.decode('utf-8') == expected_argv

    # Test Reg Average: average ref
    nr_average_4 = RegAverage()
    ref_file = example_data('anatomical.nii')
    one_file = example_data('im1.nii')
    two_file = example_data('im2.nii')
    three_file = example_data('im3.nii')
    trans1_file = example_data('roi01.nii')
    trans2_file = example_data('roi02.nii')
    trans3_file = example_data('roi03.nii')
    nr_average_4.inputs.warp_files = [
        trans1_file, one_file, trans2_file, two_file, trans3_file, three_file
    ]
    nr_average_4.inputs.avg_ref_file = ref_file
    nr_average_4.inputs.omp_core_val = 1
    generated_cmd = nr_average_4.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), 'reg_average_cmd')
    with open(reg_average_cmd, 'rb') as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = ('%s %s -avg_tran %s -omp 1 %s %s %s %s %s %s' %
                     (get_custom_path('reg_average'),
                      os.path.join(os.getcwd(), 'avg_out.nii.gz'), ref_file,
                      trans1_file, one_file, trans2_file, two_file,
                      trans3_file, three_file))

    assert argv.decode('utf-8') == expected_argv

    # Test Reg Average: demean3
    nr_average_5 = RegAverage()
    ref_file = example_data('anatomical.nii')
    one_file = example_data('im1.nii')
    two_file = example_data('im2.nii')
    three_file = example_data('im3.nii')
    aff1_file = example_data('TransformParameters.0.txt')
    aff2_file = example_data('ants_Affine.txt')
    aff3_file = example_data('elastix.txt')
    trans1_file = example_data('roi01.nii')
    trans2_file = example_data('roi02.nii')
    trans3_file = example_data('roi03.nii')
    nr_average_5.inputs.warp_files = [
        aff1_file, trans1_file, one_file, aff2_file, trans2_file, two_file,
        aff3_file, trans3_file, three_file
    ]
    nr_average_5.inputs.demean3_ref_file = ref_file
    nr_average_5.inputs.omp_core_val = 1
    generated_cmd = nr_average_5.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), 'reg_average_cmd')
    with open(reg_average_cmd, 'rb') as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = ('%s %s -demean3 %s -omp 1 %s %s %s %s %s %s %s %s %s' %
                     (get_custom_path('reg_average'),
                      os.path.join(os.getcwd(), 'avg_out.nii.gz'), ref_file,
                      aff1_file, trans1_file, one_file, aff2_file, trans2_file,
                      two_file, aff3_file, trans3_file, three_file))

    assert argv.decode('utf-8') == expected_argv


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_transform'),
    reason="niftyreg is not installed. reg_transform not found.")
def test_reg_transform_def():
    """ tests for reg_transform interface """
    # Create a reg_transform object
    nr_transform = RegTransform()

    # Check if the command is properly defined
    assert nr_transform.cmd == get_custom_path('reg_transform')

    # Assign some input data
    trans_file = example_data('warpfield.nii')
    nr_transform.inputs.def_input = trans_file
    nr_transform.inputs.omp_core_val = 4

    cmd_tmp = '{cmd} -omp 4 -def {trans_file} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        trans_file=trans_file,
        out_file=os.path.join(os.getcwd(), 'warpfield_trans.nii.gz'))

    assert nr_transform.cmdline == expected_cmd

    # Test reg_transform: def ref
    nr_transform_2 = RegTransform(omp_core_val=4)
    ref_file = example_data('im1.nii')
    trans_file = example_data('warpfield.nii')
    nr_transform_2.inputs.ref1_file = ref_file
    nr_transform_2.inputs.def_input = trans_file

    cmd_tmp = '{cmd} -ref {ref_file} -omp 4 -def {trans_file} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        ref_file=ref_file,
        trans_file=trans_file,
        out_file=os.path.join(os.getcwd(), 'warpfield_trans.nii.gz'))

    assert nr_transform_2.cmdline == expected_cmd

    # Test reg_transform: comp nii
    nr_transform_3 = RegTransform(omp_core_val=4)
    ref_file = example_data('im1.nii')
    trans_file = example_data('warpfield.nii')
    trans2_file = example_data('anatomical.nii')
    nr_transform_3.inputs.ref1_file = ref_file
    nr_transform_3.inputs.comp_input2 = trans2_file
    nr_transform_3.inputs.comp_input = trans_file

    cmd_tmp = '{cmd} -ref {ref_file} -omp 4 -comp {trans1} {trans2} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        ref_file=ref_file,
        trans1=trans_file,
        trans2=trans2_file,
        out_file=os.path.join(os.getcwd(), 'warpfield_trans.nii.gz'))

    assert nr_transform_3.cmdline == expected_cmd

    # Test reg_transform: comp txt
    nr_transform_4 = RegTransform(omp_core_val=4)
    aff1_file = example_data('ants_Affine.txt')
    aff2_file = example_data('elastix.txt')
    nr_transform_4.inputs.comp_input2 = aff2_file
    nr_transform_4.inputs.comp_input = aff1_file

    cmd_tmp = '{cmd} -omp 4 -comp {aff1} {aff2} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        aff1=aff1_file,
        aff2=aff2_file,
        out_file=os.path.join(os.getcwd(), 'ants_Affine_trans.txt'))

    assert nr_transform_4.cmdline == expected_cmd

    # Test reg_transform: comp
    nr_transform_5 = RegTransform(omp_core_val=4)
    trans_file = example_data('warpfield.nii')
    aff_file = example_data('elastix.txt')
    nr_transform_5.inputs.comp_input2 = trans_file
    nr_transform_5.inputs.comp_input = aff_file

    cmd_tmp = '{cmd} -omp 4 -comp {aff} {trans} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        aff=aff_file,
        trans=trans_file,
        out_file=os.path.join(os.getcwd(), 'elastix_trans.nii.gz'))

    assert nr_transform_5.cmdline == expected_cmd

    # Test reg_transform: flirt
    nr_transform_6 = RegTransform(omp_core_val=4)
    aff_file = example_data('elastix.txt')
    ref_file = example_data('im1.nii')
    in_file = example_data('im2.nii')
    nr_transform_6.inputs.flirt_2_nr_input = (aff_file, ref_file, in_file)

    cmd_tmp = '{cmd} -omp 4 -flirtAff2NR {aff} {ref} {in_file} {out_file}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_transform'),
        aff=aff_file,
        ref=ref_file,
        in_file=in_file,
        out_file=os.path.join(os.getcwd(), 'elastix_trans.txt'))

    assert nr_transform_6.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd='reg_measure'),
    reason="niftyreg is not installed. reg_measure not found.")
def test_reg_measure():
    """ tests for reg_measure interface """
    # Create a reg_measure object
    nr_measure = RegMeasure()

    # Check if the command is properly defined
    assert nr_measure.cmd == get_custom_path('reg_measure')

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_measure.run()

    # Assign some input data
    ref_file = example_data('im1.nii')
    flo_file = example_data('im2.nii')
    nr_measure.inputs.ref_file = ref_file
    nr_measure.inputs.flo_file = flo_file
    nr_measure.inputs.measure_type = 'lncc'
    nr_measure.inputs.omp_core_val = 4

    cmd_tmp = '{cmd} -flo {flo} -lncc -omp 4 -out {out} -ref {ref}'
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path('reg_measure'),
        flo=flo_file,
        out='im2_lncc.txt',
        ref=ref_file)

    assert nr_measure.cmdline == expected_cmd



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for the engine module
"""

from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
from copy import deepcopy
from glob import glob
import os


import pytest
from ... import engine as pe
from .test_base import EngineTestInterface


# Test graph expansion.  The following set tests the building blocks
# of the graph expansion routine.
# XXX - SG I'll create a graphical version of these tests and actually
# ensure that all connections are tested later
@pytest.mark.parametrize(
    "iterables, expected",
    [
        ({
            "1": None
        }, (1, 0)),  # test1
        ({
            "1": dict(input1=lambda: [1, 2], input2=lambda: [1, 2])
        }, (4, 0))  # test2
    ])
def test_1mod(iterables, expected):
    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    setattr(mod1, "iterables", iterables["1"])
    pipe.add_nodes([mod1])
    pipe._flatgraph = pipe._create_flat_graph()
    pipe._execgraph = pe.generate_expanded_graph(deepcopy(pipe._flatgraph))
    assert len(pipe._execgraph.nodes()) == expected[0]
    assert len(pipe._execgraph.edges()) == expected[1]


@pytest.mark.parametrize(
    "iterables, expected",
    [
        ({
            "1": {},
            "2": dict(input1=lambda: [1, 2])
        }, (3, 2)),  # test3
        ({
            "1": dict(input1=lambda: [1, 2]),
            "2": {}
        }, (4, 2)),  # test4
        ({
            "1": dict(input1=lambda: [1, 2]),
            "2": dict(input1=lambda: [1, 2])
        }, (6, 4))  # test5
    ])
def test_2mods(iterables, expected):
    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
    for nr in ["1", "2"]:
        setattr(eval("mod" + nr), "iterables", iterables[nr])
    pipe.connect([(mod1, mod2, [('output1', 'input2')])])
    pipe._flatgraph = pipe._create_flat_graph()
    pipe._execgraph = pe.generate_expanded_graph(deepcopy(pipe._flatgraph))
    assert len(pipe._execgraph.nodes()) == expected[0]
    assert len(pipe._execgraph.edges()) == expected[1]


@pytest.mark.parametrize(
    "iterables, expected, connect",
    [
        ({
            "1": {},
            "2": dict(input1=lambda: [1, 2]),
            "3": {}
        }, (5, 4), ("1-2", "2-3")),  # test6
        ({
            "1": dict(input1=lambda: [1, 2]),
            "2": {},
            "3": {}
        }, (5, 4), ("1-3", "2-3")),  # test7
        ({
            "1": dict(input1=lambda: [1, 2]),
            "2": dict(input1=lambda: [1, 2]),
            "3": {}
        }, (8, 8), ("1-3", "2-3")),  # test8
    ])
def test_3mods(iterables, expected, connect):
    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
    mod3 = pe.Node(interface=EngineTestInterface(), name='mod3')
    for nr in ["1", "2", "3"]:
        setattr(eval("mod" + nr), "iterables", iterables[nr])
    if connect == ("1-2", "2-3"):
        pipe.connect([(mod1, mod2, [('output1', 'input2')]),
                      (mod2, mod3, [('output1', 'input2')])])
    elif connect == ("1-3", "2-3"):
        pipe.connect([(mod1, mod3, [('output1', 'input1')]),
                      (mod2, mod3, [('output1', 'input2')])])
    else:
        raise Exception(
            "connect pattern is not implemented yet within the test function")
    pipe._flatgraph = pipe._create_flat_graph()
    pipe._execgraph = pe.generate_expanded_graph(deepcopy(pipe._flatgraph))
    assert len(pipe._execgraph.nodes()) == expected[0]
    assert len(pipe._execgraph.edges()) == expected[1]

    edgenum = sorted([(len(pipe._execgraph.in_edges(node)) +
                       len(pipe._execgraph.out_edges(node)))
                      for node in pipe._execgraph.nodes()])
    assert edgenum[0] > 0


def test_expansion():
    pipe1 = pe.Workflow(name='pipe1')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
    pipe1.connect([(mod1, mod2, [('output1', 'input2')])])
    pipe2 = pe.Workflow(name='pipe2')
    mod3 = pe.Node(interface=EngineTestInterface(), name='mod3')
    mod4 = pe.Node(interface=EngineTestInterface(), name='mod4')
    pipe2.connect([(mod3, mod4, [('output1', 'input2')])])
    pipe3 = pe.Workflow(name="pipe3")
    pipe3.connect([(pipe1, pipe2, [('mod2.output1', 'mod4.input1')])])
    pipe4 = pe.Workflow(name="pipe4")
    mod5 = pe.Node(interface=EngineTestInterface(), name='mod5')
    pipe4.add_nodes([mod5])
    pipe5 = pe.Workflow(name="pipe5")
    pipe5.add_nodes([pipe4])
    pipe6 = pe.Workflow(name="pipe6")
    pipe6.connect([(pipe5, pipe3, [('pipe4.mod5.output1',
                                    'pipe2.mod3.input1')])])

    pipe6._flatgraph = pipe6._create_flat_graph()


def test_iterable_expansion():
    wf1 = pe.Workflow(name='test')
    node1 = pe.Node(EngineTestInterface(), name='node1')
    node2 = pe.Node(EngineTestInterface(), name='node2')
    node1.iterables = ('input1', [1, 2])
    wf1.connect(node1, 'output1', node2, 'input2')
    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])
    wf3._flatgraph = wf3._create_flat_graph()
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 12


def test_synchronize_expansion():
    wf1 = pe.Workflow(name='test')
    node1 = pe.Node(EngineTestInterface(), name='node1')
    node1.iterables = [('input1', [1, 2]), ('input2', [3, 4, 5])]
    node1.synchronize = True
    node2 = pe.Node(EngineTestInterface(), name='node2')
    wf1.connect(node1, 'output1', node2, 'input2')
    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])
    wf3._flatgraph = wf3._create_flat_graph()
    # Each expanded graph clone has:
    # 3 node1 expansion nodes and
    # 1 node2 replicate per node1 replicate
    # => 2 * 3 = 6 nodes per expanded subgraph
    # => 18 nodes in the group
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 18


def test_synchronize_tuples_expansion():
    wf1 = pe.Workflow(name='test')

    node1 = pe.Node(EngineTestInterface(), name='node1')
    node2 = pe.Node(EngineTestInterface(), name='node2')
    node1.iterables = [('input1', 'input2'), [(1, 3), (2, 4), (None, 5)]]

    node1.synchronize = True

    wf1.connect(node1, 'output1', node2, 'input2')

    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])

    wf3._flatgraph = wf3._create_flat_graph()
    # Identical to test_synchronize_expansion
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 18


def test_itersource_expansion():

    wf1 = pe.Workflow(name='test')
    node1 = pe.Node(EngineTestInterface(), name='node1')
    node1.iterables = ('input1', [1, 2])

    node2 = pe.Node(EngineTestInterface(), name='node2')
    wf1.connect(node1, 'output1', node2, 'input1')

    node3 = pe.Node(EngineTestInterface(), name='node3')
    node3.itersource = ('node1', 'input1')
    node3.iterables = [('input1', {1: [3, 4], 2: [5, 6, 7]})]

    wf1.connect(node2, 'output1', node3, 'input1')
    node4 = pe.Node(EngineTestInterface(), name='node4')

    wf1.connect(node3, 'output1', node4, 'input1')

    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])

    wf3._flatgraph = wf3._create_flat_graph()

    # each expanded graph clone has:
    # 2 node1 expansion nodes,
    # 1 node2 per node1 replicate,
    # 2 node3 replicates for the node1 input1 value 1,
    # 3 node3 replicates for the node1 input1 value 2 and
    # 1 node4 successor per node3 replicate
    # => 2 + 2 + (2 + 3) + 5 = 14 nodes per expanded graph clone
    # => 3 * 14 = 42 nodes in the group
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 42


def test_itersource_synchronize1_expansion():
    wf1 = pe.Workflow(name='test')
    node1 = pe.Node(EngineTestInterface(), name='node1')
    node1.iterables = [('input1', [1, 2]), ('input2', [3, 4])]
    node1.synchronize = True
    node2 = pe.Node(EngineTestInterface(), name='node2')
    wf1.connect(node1, 'output1', node2, 'input1')
    node3 = pe.Node(EngineTestInterface(), name='node3')
    node3.itersource = ('node1', ['input1', 'input2'])
    node3.iterables = [('input1', {
        (1, 3): [5, 6]
    }), ('input2', {
        (1, 3): [7, 8],
        (2, 4): [9]
    })]
    wf1.connect(node2, 'output1', node3, 'input1')
    node4 = pe.Node(EngineTestInterface(), name='node4')
    wf1.connect(node3, 'output1', node4, 'input1')
    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])
    wf3._flatgraph = wf3._create_flat_graph()

    # each expanded graph clone has:
    # 2 node1 expansion nodes,
    # 1 node2 per node1 replicate,
    # 2 node3 replicates for the node1 input1 value 1,
    # 3 node3 replicates for the node1 input1 value 2 and
    # 1 node4 successor per node3 replicate
    # => 2 + 2 + (2 + 3) + 5 = 14 nodes per expanded graph clone
    # => 3 * 14 = 42 nodes in the group
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 42


def test_itersource_synchronize2_expansion():
    wf1 = pe.Workflow(name='test')

    node1 = pe.Node(EngineTestInterface(), name='node1')
    node1.iterables = [('input1', [1, 2]), ('input2', [3, 4])]
    node1.synchronize = True
    node2 = pe.Node(EngineTestInterface(), name='node2')
    wf1.connect(node1, 'output1', node2, 'input1')
    node3 = pe.Node(EngineTestInterface(), name='node3')
    node3.itersource = ('node1', ['input1', 'input2'])
    node3.synchronize = True
    node3.iterables = [('input1', 'input2'), {
        (1, 3): [(5, 7), (6, 8)],
        (2, 4): [(None, 9)]
    }]
    wf1.connect(node2, 'output1', node3, 'input1')
    node4 = pe.Node(EngineTestInterface(), name='node4')
    wf1.connect(node3, 'output1', node4, 'input1')
    wf3 = pe.Workflow(name='group')
    for i in [0, 1, 2]:
        wf3.add_nodes([wf1.clone(name='test%d' % i)])
    wf3._flatgraph = wf3._create_flat_graph()

    # each expanded graph clone has:
    # 2 node1 expansion nodes,
    # 1 node2 per node1 replicate,
    # 2 node3 replicates for the node1 input1 value 1,
    # 1 node3 replicates for the node1 input1 value 2 and
    # 1 node4 successor per node3 replicate
    # => 2 + 2 + (2 + 1) + 3 = 10 nodes per expanded graph clone
    # => 3 * 10 = 30 nodes in the group
    assert len(pe.generate_expanded_graph(wf3._flatgraph).nodes()) == 30



def test_old_config(tmpdir):
    tmpdir.chdir()
    wd = os.getcwd()
    from nipype.interfaces.utility import Function

    def func1():
        return 1

    def func2(a):
        return a + 1

    n1 = pe.Node(
        Function(input_names=[], output_names=['a'], function=func1),
        name='n1')
    n2 = pe.Node(
        Function(input_names=['a'], output_names=['b'], function=func2),
        name='n2')
    w1 = pe.Workflow(name='test')
    modify = lambda x: x + 1
    n1.inputs.a = 1
    w1.connect(n1, ('a', modify), n2, 'a')
    w1.base_dir = wd

    w1.config['execution']['crashdump_dir'] = wd
    # generate outputs

    w1.run(plugin='Linear')


def test_mapnode_json(tmpdir):
    """Tests that mapnodes don't generate excess jsons
    """
    tmpdir.chdir()
    wd = os.getcwd()
    from nipype import MapNode, Function, Workflow

    def func1(in1):
        return in1 + 1

    n1 = MapNode(
        Function(input_names=['in1'], output_names=['out'], function=func1),
        iterfield=['in1'],
        name='n1')
    n1.inputs.in1 = [1]
    w1 = Workflow(name='test')
    w1.base_dir = wd
    w1.config['execution']['crashdump_dir'] = wd
    w1.add_nodes([n1])
    w1.run()
    n1.inputs.in1 = [2]
    w1.run()
    # should rerun
    n1.inputs.in1 = [1]
    eg = w1.run()

    node = list(eg.nodes())[0]
    outjson = glob(os.path.join(node.output_dir(), '_0x*.json'))
    assert len(outjson) == 1

    # check that multiple json's don't trigger rerun
    with open(os.path.join(node.output_dir(), 'test.json'), 'wt') as fp:
        fp.write('dummy file')
    w1.config['execution'].update(**{'stop_on_first_rerun': True})

    w1.run()


def test_parameterize_dirs_false(tmpdir):
    from ....interfaces.utility import IdentityInterface
    from ....testing import example_data

    input_file = example_data('fsl_motion_outliers_fd.txt')

    n1 = pe.Node(EngineTestInterface(), name='Node1')
    n1.iterables = ('input_file', (input_file, input_file))
    n1.interface.inputs.input1 = 1

    n2 = pe.Node(IdentityInterface(fields='in1'), name='Node2')

    wf = pe.Workflow(name='Test')
    wf.base_dir = tmpdir.strpath
    wf.config['execution']['parameterize_dirs'] = False
    wf.connect([(n1, n2, [('output1', 'in1')])])

    wf.run()


def test_serial_input(tmpdir):
    tmpdir.chdir()
    wd = os.getcwd()
    from nipype import MapNode, Function, Workflow

    def func1(in1):
        return in1

    n1 = MapNode(
        Function(input_names=['in1'], output_names=['out'], function=func1),
        iterfield=['in1'],
        name='n1')
    n1.inputs.in1 = [1, 2, 3]

    w1 = Workflow(name='test')
    w1.base_dir = wd
    w1.add_nodes([n1])
    # set local check
    w1.config['execution'] = {
        'stop_on_first_crash': 'true',
        'local_hash_check': 'true',
        'crashdump_dir': wd,
        'poll_sleep_duration': 2
    }

    # test output of num_subnodes method when serial is default (False)
    assert n1.num_subnodes() == len(n1.inputs.in1)

    # test running the workflow on default conditions
    w1.run(plugin='MultiProc')

    # test output of num_subnodes method when serial is True
    n1._serial = True
    assert n1.num_subnodes() == 1

    # test running the workflow on serial conditions
    w1.run(plugin='MultiProc')


def test_write_graph_runs(tmpdir):
    tmpdir.chdir()

    for graph in ('orig', 'flat', 'exec', 'hierarchical', 'colored'):
        for simple in (True, False):
            pipe = pe.Workflow(name='pipe')
            mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
            mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
            pipe.connect([(mod1, mod2, [('output1', 'input1')])])
            try:
                pipe.write_graph(
                    graph2use=graph, simple_form=simple, format='dot')
            except Exception:
                assert False, \
                    'Failed to plot {} {} graph'.format(
                        'simple' if simple else 'detailed', graph)

            assert os.path.exists('graph.dot') or os.path.exists(
                'graph_detailed.dot')

            try:
                os.remove('graph.dot')
            except OSError:
                pass
            try:
                os.remove('graph_detailed.dot')
            except OSError:
                pass


def test_deep_nested_write_graph_runs(tmpdir):
    tmpdir.chdir()

    for graph in ('orig', 'flat', 'exec', 'hierarchical', 'colored'):
        for simple in (True, False):
            pipe = pe.Workflow(name='pipe')
            parent = pipe
            for depth in range(10):
                sub = pe.Workflow(name='pipe_nest_{}'.format(depth))
                parent.add_nodes([sub])
                parent = sub
            mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
            parent.add_nodes([mod1])
            try:
                pipe.write_graph(
                    graph2use=graph, simple_form=simple, format='dot')
            except Exception as e:
                assert False, \
                    'Failed to plot {} {} deep graph: {!s}'.format(
                        'simple' if simple else 'detailed', graph, e)

            assert os.path.exists('graph.dot') or os.path.exists(
                'graph_detailed.dot')
            try:
                os.remove('graph.dot')
            except OSError:
                pass
            try:
                os.remove('graph_detailed.dot')
            except OSError:
                pass

import networkx
# Format of the graph has slightly changed
graph_str = '""' if int(networkx.__version__.split('.')[0]) == 1 else ''

# examples of dot files used in the following test
dotfile_orig = ['strict digraph ' + graph_str + ' {\n',
                '"mod1 (engine)";\n',
                '"mod2 (engine)";\n',
                '"mod1 (engine)" -> "mod2 (engine)";\n',
                '}\n']

dotfile_detailed_orig = ['digraph structs {\n',
                         'node [shape=record];\n',
                         'pipemod1 [label="{IN}|{ mod1 | engine |  }|{OUT|<outoutput1> output1}"];\n',
                         'pipemod2 [label="{IN|<ininput1> input1}|{ mod2 | engine |  }|{OUT}"];\n',
                         'pipemod1:outoutput1:e -> pipemod2:ininput1:w;\n',
                         '}']


dotfile_hierarchical = ['digraph pipe{\n',
                        '  label="pipe";\n',
                        '  pipe_mod1[label="mod1 (engine)"];\n',
                        '  pipe_mod2[label="mod2 (engine)"];\n',
                        '  pipe_mod1 -> pipe_mod2;\n',
                        '}']

dotfile_colored = ['digraph pipe{\n',
                   '  label="pipe";\n',
                   '  pipe_mod1[label="mod1 (engine)", style=filled, fillcolor="#FFFFC8"];\n',
                   '  pipe_mod2[label="mod2 (engine)", style=filled, fillcolor="#FFFFC8"];\n',
                   '  pipe_mod1 -> pipe_mod2;\n',
                   '}']

dotfiles = {
    "orig": dotfile_orig,
    "flat": dotfile_orig,
    "exec": dotfile_orig,
    "hierarchical": dotfile_hierarchical,
    "colored": dotfile_colored
    }

@pytest.mark.parametrize("simple", [True, False])
@pytest.mark.parametrize("graph_type", ['orig', 'flat', 'exec', 'hierarchical', 'colored'])
def test_write_graph_dotfile(tmpdir, graph_type, simple):
    """ checking dot files for a workflow without iterables"""
    tmpdir.chdir()

    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
    pipe.connect([(mod1, mod2, [('output1', 'input1')])])
    pipe.write_graph(
        graph2use=graph_type, simple_form=simple, format='dot')

    with open("graph.dot") as f:
        graph_str = f.read()

    if simple:
        for line in dotfiles[graph_type]:
            assert line in graph_str
    else:
        # if simple=False graph.dot uses longer names
        for line in dotfiles[graph_type]:
            if graph_type in ["hierarchical", "colored"]:
                assert line.replace("mod1 (engine)", "mod1.EngineTestInterface.engine").replace(
                    "mod2 (engine)", "mod2.EngineTestInterface.engine") in graph_str
            else:
                assert line.replace(
                    "mod1 (engine)", "pipe.mod1.EngineTestInterface.engine").replace(
                    "mod2 (engine)", "pipe.mod2.EngineTestInterface.engine") in graph_str

    # graph_detailed is the same for orig, flat, exec (if no iterables)
    # graph_detailed is not created for hierachical or colored
    if graph_type not in ["hierarchical", "colored"]:
        with open("graph_detailed.dot") as f:
            graph_str = f.read()
        for line in dotfile_detailed_orig:
            assert line in graph_str


# examples of dot files used in the following test
dotfile_detailed_iter_exec = [
    'digraph structs {\n',
    'node [shape=record];\n',
    'pipemod1aIa1 [label="{IN}|{ a1 | engine | mod1.aI }|{OUT|<outoutput1> output1}"];\n',
    'pipemod2a1 [label="{IN|<ininput1> input1}|{ a1 | engine | mod2 }|{OUT}"];\n',
    'pipemod1aIa0 [label="{IN}|{ a0 | engine | mod1.aI }|{OUT|<outoutput1> output1}"];\n',
    'pipemod2a0 [label="{IN|<ininput1> input1}|{ a0 | engine | mod2 }|{OUT}"];\n',
    'pipemod1aIa0:outoutput1:e -> pipemod2a0:ininput1:w;\n',
    'pipemod1aIa1:outoutput1:e -> pipemod2a1:ininput1:w;\n',
    '}']

dotfile_iter_hierarchical = [
    'digraph pipe{\n',
    '  label="pipe";\n',
    '  pipe_mod1[label="mod1 (engine)", shape=box3d,style=filled, color=black, colorscheme=greys7 fillcolor=2];\n',
    '  pipe_mod2[label="mod2 (engine)"];\n',
    '  pipe_mod1 -> pipe_mod2;\n',
    '}']

dotfile_iter_colored = [
    'digraph pipe{\n',
    '  label="pipe";\n',
    '  pipe_mod1[label="mod1 (engine)", shape=box3d,style=filled, color=black, colorscheme=greys7 fillcolor=2];\n',
    '  pipe_mod2[label="mod2 (engine)", style=filled, fillcolor="#FFFFC8"];\n',
    '  pipe_mod1 -> pipe_mod2;\n',
    '}']

dotfiles_iter = {
    "orig": dotfile_orig,
    "flat": dotfile_orig,
    "exec": dotfile_orig,
    "hierarchical": dotfile_iter_hierarchical,
    "colored": dotfile_iter_colored
    }

dotfiles_detailed_iter = {
    "orig": dotfile_detailed_orig,
    "flat": dotfile_detailed_orig,
    "exec": dotfile_detailed_iter_exec
    }

@pytest.mark.parametrize("simple", [True, False])
@pytest.mark.parametrize("graph_type", ['orig', 'flat', 'exec', 'hierarchical', 'colored'])
def test_write_graph_dotfile_iterables(tmpdir, graph_type, simple):
    """ checking dot files for a workflow with iterables"""
    tmpdir.chdir()

    pipe = pe.Workflow(name='pipe')
    mod1 = pe.Node(interface=EngineTestInterface(), name='mod1')
    mod1.iterables = ('input1', [1, 2])
    mod2 = pe.Node(interface=EngineTestInterface(), name='mod2')
    pipe.connect([(mod1, mod2, [('output1', 'input1')])])
    pipe.write_graph(
        graph2use=graph_type, simple_form=simple, format='dot')

    with open("graph.dot") as f:
        graph_str = f.read()

    if simple:
        for line in dotfiles_iter[graph_type]:
            assert line in graph_str
    else:
        # if simple=False graph.dot uses longer names
        for line in dotfiles_iter[graph_type]:
            if graph_type in ["hierarchical", "colored"]:
                assert line.replace("mod1 (engine)", "mod1.EngineTestInterface.engine").replace(
                    "mod2 (engine)", "mod2.EngineTestInterface.engine") in graph_str
            else:
                assert line.replace(
                    "mod1 (engine)", "pipe.mod1.EngineTestInterface.engine").replace(
                    "mod2 (engine)", "pipe.mod2.EngineTestInterface.engine") in graph_str

    # graph_detailed is not created for hierachical or colored
    if graph_type not in ["hierarchical", "colored"]:
        with open("graph_detailed.dot") as f:
            graph_str = f.read()
        for line in dotfiles_detailed_iter[graph_type]:
            assert line in graph_str



def test_io_subclass():
    """Ensure any io subclass allows dynamic traits"""
    from nipype.interfaces.io import IOBase
    from nipype.interfaces.base import DynamicTraitedSpec

    class TestKV(IOBase):
        _always_run = True
        output_spec = DynamicTraitedSpec

        def _list_outputs(self):
            outputs = {}
            outputs['test'] = 1
            outputs['foo'] = 'bar'
            return outputs

    wf = pe.Workflow('testkv')

    def testx2(test):
        return test * 2

    kvnode = pe.Node(TestKV(), name='testkv')
    from nipype.interfaces.utility import Function
    func = pe.Node(
        Function(
            input_names=['test'], output_names=['test2'], function=testx2),
        name='func')
    exception_not_raised = True
    try:
        wf.connect(kvnode, 'test', func, 'test')
    except Exception as e:
        if 'Module testkv has no output called test' in e:
            exception_not_raised = False
    assert exception_not_raised



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import NwarpApply


def test_NwarpApply_inputs():
    input_map = dict(
        ainterp=dict(argstr='-ainterp %s', ),
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='-source %s',
            mandatory=True,
        ),
        interp=dict(
            argstr='-interp %s',
            usedefault=True,
        ),
        inv_warp=dict(argstr='-iwarp', ),
        master=dict(
            argstr='-master %s',
            extensions=None,
        ),
        out_file=dict(
            argstr='-prefix %s',
            extensions=None,
            name_source='in_file',
            name_template='%s_Nwarp',
        ),
        quiet=dict(
            argstr='-quiet',
            xor=['verb'],
        ),
        short=dict(argstr='-short', ),
        verb=dict(
            argstr='-verb',
            xor=['quiet'],
        ),
        warp=dict(
            argstr='-nwarp %s',
            mandatory=True,
        ),
    )
    inputs = NwarpApply.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_NwarpApply_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = NwarpApply.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..dti import TractSkeleton


def test_TractSkeleton_inputs():
    input_map = dict(
        alt_data_file=dict(
            argstr='-a %s',
            extensions=None,
        ),
        alt_skeleton=dict(
            argstr='-s %s',
            extensions=None,
        ),
        args=dict(argstr='%s', ),
        data_file=dict(extensions=None, ),
        distance_map=dict(extensions=None, ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='-i %s',
            extensions=None,
            mandatory=True,
        ),
        output_type=dict(),
        project_data=dict(
            argstr='-p %.3f %s %s %s %s',
            requires=['threshold', 'distance_map', 'data_file'],
        ),
        projected_data=dict(extensions=None, ),
        search_mask_file=dict(
            extensions=None,
            xor=['use_cingulum_mask'],
        ),
        skeleton_file=dict(argstr='-o %s', ),
        threshold=dict(),
        use_cingulum_mask=dict(
            usedefault=True,
            xor=['search_mask_file'],
        ),
    )
    inputs = TractSkeleton.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_TractSkeleton_outputs():
    output_map = dict(
        projected_data=dict(extensions=None, ),
        skeleton_file=dict(extensions=None, ),
    )
    outputs = TractSkeleton.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import Edge3


def test_Edge3_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        datum=dict(argstr='-datum %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fscale=dict(
            argstr='-fscale',
            xor=['gscale', 'nscale', 'scale_floats'],
        ),
        gscale=dict(
            argstr='-gscale',
            xor=['fscale', 'nscale', 'scale_floats'],
        ),
        in_file=dict(
            argstr='-input %s',
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=0,
        ),
        nscale=dict(
            argstr='-nscale',
            xor=['fscale', 'gscale', 'scale_floats'],
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr='-prefix %s',
            extensions=None,
            position=-1,
        ),
        outputtype=dict(),
        scale_floats=dict(
            argstr='-scale_floats %f',
            xor=['fscale', 'gscale', 'nscale'],
        ),
        verbose=dict(argstr='-verbose', ),
    )
    inputs = Edge3.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Edge3_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = Edge3.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""
Utilities for the CLI functions.
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import bytes, str

import re
import click
import json

from .instance import import_module
from ..interfaces.base import InputMultiPath, traits
from ..interfaces.base.support import get_trait_desc

# different context options
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
UNKNOWN_OPTIONS = dict(allow_extra_args=True, ignore_unknown_options=True)

# specification of existing ParamTypes
ExistingDirPath = click.Path(exists=True, file_okay=False, resolve_path=True)
ExistingFilePath = click.Path(exists=True, dir_okay=False, resolve_path=True)
UnexistingFilePath = click.Path(dir_okay=False, resolve_path=True)


# validators
def check_not_none(ctx, param, value):
    if value is None:
        raise click.BadParameter('got {}.'.format(value))
    return value


# declare custom click.ParamType
class RegularExpression(click.ParamType):
    name = 'regex'

    def convert(self, value, param, ctx):
        try:
            rex = re.compile(value, re.IGNORECASE)
        except ValueError:
            self.fail('%s is not a valid regular expression.' % value, param,
                      ctx)
        else:
            return rex


class PythonModule(click.ParamType):
    name = 'Python module path'

    def convert(self, value, param, ctx):
        try:
            module = import_module(value)
        except ValueError:
            self.fail('%s is not a valid Python module.' % value, param, ctx)
        else:
            return module


def add_args_options(arg_parser, interface):
    """Add arguments to `arg_parser` to create a CLI for `interface`."""
    inputs = interface.input_spec()
    for name, spec in sorted(interface.inputs.traits(transient=None).items()):
        desc = "\n".join(get_trait_desc(inputs, name, spec))[len(name) + 2:]
        # Escape any % signs with a %
        desc = desc.replace('%', '%%')
        args = {}
        has_multiple_inner_traits = False

        if spec.is_trait_type(traits.Bool):
            args["default"] = getattr(inputs, name)
            args["action"] = 'store_true'

        # current support is for simple trait types
        if not spec.inner_traits:
            if not spec.is_trait_type(traits.TraitCompound):
                trait_type = type(spec.trait_type.default_value)
            if trait_type in (bytes, str, int, float):
                if trait_type == bytes:
                    trait_type = str
                args["type"] = trait_type
        elif len(spec.inner_traits) == 1:
            trait_type = type(spec.inner_traits[0].trait_type.default_value)
            if trait_type == bytes:
                trait_type = str
            if trait_type in (bytes, bool, str, int, float):
                args["type"] = trait_type
        else:
            if len(spec.inner_traits) > 1:
                if not spec.is_trait_type(traits.Dict):
                    has_multiple_inner_traits = True

        if getattr(spec, "mandatory", False):
            if spec.is_trait_type(InputMultiPath):
                args["nargs"] = "+"
            elif spec.is_trait_type(traits.List):
                if (spec.trait_type.minlen == spec.trait_type.maxlen) and \
                        spec.trait_type.maxlen:
                    args["nargs"] = spec.trait_type.maxlen
                else:
                    args["nargs"] = "+"
            elif spec.is_trait_type(traits.Dict):
                args["type"] = json.loads

            if has_multiple_inner_traits:
                raise NotImplementedError(
                    ('This interface cannot be used. via the'
                     ' command line as multiple inner traits'
                     ' are currently not supported for mandatory'
                     ' argument: {}.'.format(name)))
            arg_parser.add_argument(name, help=desc, **args)
        else:
            if spec.is_trait_type(InputMultiPath):
                args["nargs"] = "*"
            elif spec.is_trait_type(traits.List):
                if (spec.trait_type.minlen == spec.trait_type.maxlen) and \
                        spec.trait_type.maxlen:
                    args["nargs"] = spec.trait_type.maxlen
                else:
                    args["nargs"] = "*"
            if not has_multiple_inner_traits:
                arg_parser.add_argument(
                    "--%s" % name, dest=name, help=desc, **args)

    return arg_parser



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import CommandLine, CommandLineInputSpec, SEMLikeCommandLine, TraitedSpec, File, Directory, traits, isdefined, InputMultiPath, OutputMultiPath
import os


class DicomToNrrdConverterInputSpec(CommandLineInputSpec):
    inputDicomDirectory = Directory(
        desc="Directory holding Dicom series",
        exists=True,
        argstr="--inputDicomDirectory %s")
    outputDirectory = traits.Either(
        traits.Bool,
        Directory(),
        hash_files=False,
        desc="Directory holding the output NRRD format",
        argstr="--outputDirectory %s")
    outputVolume = traits.Str(
        desc="Output filename (.nhdr or .nrrd)", argstr="--outputVolume %s")
    smallGradientThreshold = traits.Float(
        desc=
        "If a gradient magnitude is greater than 0 and less than smallGradientThreshold, then DicomToNrrdConverter will display an error message and quit, unless the useBMatrixGradientDirections option is set.",
        argstr="--smallGradientThreshold %f")
    writeProtocolGradientsFile = traits.Bool(
        desc=
        "Write the protocol gradients to a file suffixed by \'.txt\' as they were specified in the procol by multiplying each diffusion gradient direction by the measurement frame.  This file is for debugging purposes only, the format is not fixed, and will likely change as debugging of new dicom formats is necessary.",
        argstr="--writeProtocolGradientsFile ")
    useIdentityMeaseurementFrame = traits.Bool(
        desc=
        "Adjust all the gradients so that the measurement frame is an identity matrix.",
        argstr="--useIdentityMeaseurementFrame ")
    useBMatrixGradientDirections = traits.Bool(
        desc=
        "Fill the nhdr header with the gradient directions and bvalues computed out of the BMatrix. Only changes behavior for Siemens data.",
        argstr="--useBMatrixGradientDirections ")


class DicomToNrrdConverterOutputSpec(TraitedSpec):
    outputDirectory = Directory(
        desc="Directory holding the output NRRD format", exists=True)


class DicomToNrrdConverter(SEMLikeCommandLine):
    """title: DICOM to NRRD Converter

category: Converters

description: Converts diffusion weighted MR images in dicom series into Nrrd format for analysis in Slicer. This program has been tested on only a limited subset of DTI dicom formats available from Siemens, GE, and Phillips scanners. Work in progress to support dicom multi-frame data. The program parses dicom header to extract necessary information about measurement frame, diffusion weighting directions, b-values, etc, and write out a nrrd image. For non-diffusion weighted dicom images, it loads in an entire dicom series and writes out a single dicom volume in a .nhdr/.raw pair.

version: 0.2.0.$Revision: 916 $(alpha)

documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/DicomToNrrdConverter

license: https://www.nitrc.org/svn/brains/BuildScripts/trunk/License.txt

contributor: Xiaodong Tao (GE), Vince Magnotta (UIowa), Hans Johnson (UIowa)

acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health through the NIH Roadmap for Medical Research, Grant U54 EB005149.  Additional support for DTI data produced on Philips scanners was contributed by Vincent Magnotta and Hans Johnson at the University of Iowa.

"""

    input_spec = DicomToNrrdConverterInputSpec
    output_spec = DicomToNrrdConverterOutputSpec
    _cmd = "DicomToNrrdConverter "
    _outputs_filenames = {'outputDirectory': 'outputDirectory'}


class OrientScalarVolumeInputSpec(CommandLineInputSpec):
    inputVolume1 = File(
        position=-2, desc="Input volume 1", exists=True, argstr="%s")
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="The oriented volume",
        argstr="%s")
    orientation = traits.Enum(
        "Axial",
        "Coronal",
        "Sagittal",
        "RIP",
        "LIP",
        "RSP",
        "LSP",
        "RIA",
        "LIA",
        "RSA",
        "LSA",
        "IRP",
        "ILP",
        "SRP",
        "SLP",
        "IRA",
        "ILA",
        "SRA",
        "SLA",
        "RPI",
        "LPI",
        "RAI",
        "LAI",
        "RPS",
        "LPS",
        "RAS",
        "LAS",
        "PRI",
        "PLI",
        "ARI",
        "ALI",
        "PRS",
        "PLS",
        "ARS",
        "ALS",
        "IPR",
        "SPR",
        "IAR",
        "SAR",
        "IPL",
        "SPL",
        "IAL",
        "SAL",
        "PIR",
        "PSR",
        "AIR",
        "ASR",
        "PIL",
        "PSL",
        "AIL",
        "ASL",
        desc="Orientation choices",
        argstr="--orientation %s")


class OrientScalarVolumeOutputSpec(TraitedSpec):
    outputVolume = File(position=-1, desc="The oriented volume", exists=True)


class OrientScalarVolume(SEMLikeCommandLine):
    """title: Orient Scalar Volume

category: Converters

description: Orients an output volume. Rearranges the slices in a volume according to the selected orientation. The slices are not interpolated. They are just reordered and/or permuted. The resulting volume will cover the original volume. NOTE: since Slicer takes into account the orientation of a volume, the re-oriented volume will not show any difference from the original volume, To see the difference, save the volume and display it with a system that either ignores the orientation of the image (e.g. Paraview) or displays individual images.

version: 0.1.0.$Revision: 19608 $(alpha)

documentation-url: http://wiki.slicer.org/slicerWiki/index.php/Documentation/4.1/Modules/OrientImage

contributor: Bill Lorensen (GE)

acknowledgements: This work is part of the National Alliance for Medical Image Computing (NAMIC), funded by the National Institutes of Health through the NIH Roadmap for Medical Research, Grant U54 EB005149.

"""

    input_spec = OrientScalarVolumeInputSpec
    output_spec = OrientScalarVolumeOutputSpec
    _cmd = "OrientScalarVolume "
    _outputs_filenames = {'outputVolume': 'outputVolume.nii'}



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

import os

from ....base import (CommandLine, CommandLineInputSpec, SEMLikeCommandLine,
                      TraitedSpec, File, Directory, traits, isdefined,
                      InputMultiPath, OutputMultiPath)


class fiberprocessInputSpec(CommandLineInputSpec):
    fiber_file = File(
        desc="DTI fiber file", exists=True, argstr="--fiber_file %s")
    fiber_output = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=
        "Output fiber file. May be warped or updated with new data depending on other options used.",
        argstr="--fiber_output %s")
    tensor_volume = File(
        desc="Interpolate tensor values from the given field",
        exists=True,
        argstr="--tensor_volume %s")
    h_field = File(
        desc=
        "HField for warp and statistics lookup. If this option is used tensor-volume must also be specified.",
        exists=True,
        argstr="--h_field %s")
    displacement_field = File(
        desc=
        "Displacement Field for warp and statistics lookup.  If this option is used tensor-volume must also be specified.",
        exists=True,
        argstr="--displacement_field %s")
    saveProperties = traits.Bool(
        desc=
        "save the tensor property as scalar data into the vtk (only works for vtk fiber files). ",
        argstr="--saveProperties ")
    no_warp = traits.Bool(
        desc=
        "Do not warp the geometry of the tensors only obtain the new statistics.",
        argstr="--no_warp ")
    fiber_radius = traits.Float(
        desc="set radius of all fibers to this value",
        argstr="--fiber_radius %f")
    index_space = traits.Bool(
        desc=
        "Use index-space for fiber output coordinates, otherwise us world space for fiber output coordinates (from tensor file).",
        argstr="--index_space ")
    voxelize = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc=
        "Voxelize fiber into a label map (the labelmap filename is the argument of -V). The tensor file must be specified using -T for information about the size, origin, spacing of the image. The deformation is applied before the voxelization ",
        argstr="--voxelize %s")
    voxelize_count_fibers = traits.Bool(
        desc="Count number of fibers per-voxel instead of just setting to 1",
        argstr="--voxelize_count_fibers ")
    voxel_label = traits.Int(
        desc="Label for voxelized fiber", argstr="--voxel_label %d")
    verbose = traits.Bool(desc="produce verbose output", argstr="--verbose ")
    noDataChange = traits.Bool(
        desc="Do not change data ??? ", argstr="--noDataChange ")


class fiberprocessOutputSpec(TraitedSpec):
    fiber_output = File(
        desc=
        "Output fiber file. May be warped or updated with new data depending on other options used.",
        exists=True)
    voxelize = File(
        desc=
        "Voxelize fiber into a label map (the labelmap filename is the argument of -V). The tensor file must be specified using -T for information about the size, origin, spacing of the image. The deformation is applied before the voxelization ",
        exists=True)


class fiberprocess(SEMLikeCommandLine):
    """title: FiberProcess (DTIProcess)

category: Diffusion.Tractography

description: fiberprocess is a tool that manage fiber files extracted from the fibertrack tool or any fiber tracking algorithm. It takes as an input .fib and .vtk files (--fiber_file) and saves the changed fibers (--fiber_output) into the 2 same formats. The main purpose of this tool is to deform the fiber file with a transformation field as an input (--displacement_field or --h_field depending if you deal with dfield or hfield). To use that option you need to specify the tensor field from which the fiber file was extracted with the option --tensor_volume. The transformation applied on the fiber file is the inverse of the one input. If the transformation is from one case to an atlas, fiberprocess assumes that the fiber file is in the atlas space and you want it in the original case space, so it's the inverse of the transformation which has been computed.
You have 2 options for fiber modification. You can either deform the fibers (their geometry) into the space OR you can keep the same geometry but map the diffusion properties (fa, md, lbd's...) of the original tensor field along the fibers at the corresponding locations. This is triggered by the --no_warp option. To use the previous example: when you have a tensor field in the original space and the deformed tensor field in the atlas space, you want to track the fibers in the atlas space, keeping this geometry but with the original case diffusion properties. Then you can specify the transformations field (from original case -> atlas) and the original tensor field with the --tensor_volume option.
With fiberprocess you can also binarize a fiber file. Using the --voxelize option will create an image where each voxel through which a fiber is passing is set to 1. The output is going to be a binary image with the values 0 or 1 by default but the 1 value voxel can be set to any number with the --voxel_label option. Finally you can create an image where the value at the voxel is the number of fiber passing through. (--voxelize_count_fibers)

version: 1.0.0

documentation-url: http://www.slicer.org/slicerWiki/index.php/Documentation/Nightly/Extensions/DTIProcess

license: Copyright (c)  Casey Goodlett. All rights reserved.
    See http://www.ia.unc.edu/dev/Copyright.htm for details.
    This software is distributed WITHOUT ANY WARRANTY; without even
    the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
    PURPOSE.  See the above copyright notices for more information.

contributor: Casey Goodlett

"""

    input_spec = fiberprocessInputSpec
    output_spec = fiberprocessOutputSpec
    _cmd = " fiberprocess "
    _outputs_filenames = {
        'fiber_output': 'fiber_output.vtk',
        'voxelize': 'voxelize.nii'
    }
    _redirect_x = False



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import OutlierCount


def test_OutlierCount_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        autoclip=dict(
            argstr='-autoclip',
            usedefault=True,
            xor=['mask'],
        ),
        automask=dict(
            argstr='-automask',
            usedefault=True,
            xor=['mask'],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fraction=dict(
            argstr='-fraction',
            usedefault=True,
        ),
        in_file=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        interval=dict(
            argstr='-range',
            usedefault=True,
        ),
        legendre=dict(
            argstr='-legendre',
            usedefault=True,
        ),
        mask=dict(
            argstr='-mask %s',
            extensions=None,
            xor=['autoclip', 'automask'],
        ),
        out_file=dict(
            extensions=None,
            keep_extension=False,
            name_source=['in_file'],
            name_template='%s_outliers',
        ),
        outliers_file=dict(
            argstr='-save %s',
            extensions=None,
            keep_extension=True,
            name_source=['in_file'],
            name_template='%s_outliers',
            output_name='out_outliers',
        ),
        polort=dict(argstr='-polort %d', ),
        qthr=dict(
            argstr='-qthr %.5f',
            usedefault=True,
        ),
        save_outliers=dict(usedefault=True, ),
    )
    inputs = OutlierCount.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_OutlierCount_outputs():
    output_map = dict(
        out_file=dict(extensions=None, ),
        out_outliers=dict(extensions=None, ),
    )
    outputs = OutlierCount.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..developer import JistBrainMp2rageDuraEstimation


def test_JistBrainMp2rageDuraEstimation_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inDistance=dict(argstr='--inDistance %f', ),
        inSecond=dict(
            argstr='--inSecond %s',
            extensions=None,
        ),
        inSkull=dict(
            argstr='--inSkull %s',
            extensions=None,
        ),
        inoutput=dict(argstr='--inoutput %s', ),
        null=dict(argstr='--null %s', ),
        outDura=dict(
            argstr='--outDura %s',
            hash_files=False,
        ),
        xDefaultMem=dict(argstr='-xDefaultMem %d', ),
        xMaxProcess=dict(
            argstr='-xMaxProcess %d',
            usedefault=True,
        ),
        xPrefExt=dict(argstr='--xPrefExt %s', ),
    )
    inputs = JistBrainMp2rageDuraEstimation.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_JistBrainMp2rageDuraEstimation_outputs():
    output_map = dict(outDura=dict(extensions=None, ), )
    outputs = JistBrainMp2rageDuraEstimation.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
